/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.farm;

import nl.cloudfarming.client.model.FarmID;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;
import nl.cloudfarming.client.farm.model.Farm;
import nl.cloudfarming.client.model.EntityCache;
import nl.cloudfarming.client.model.EntitySavable;
import org.netbeans.api.actions.Savable;
import org.netbeans.api.project.Project;
import org.netbeans.api.project.ProjectInformation;
import org.netbeans.api.project.ui.OpenProjects;
import org.netbeans.spi.project.*;
import org.netbeans.spi.project.support.ProjectOperations;
import org.netbeans.spi.project.ui.ProjectOpenedHook;
import org.netbeans.spi.project.ui.support.DefaultProjectOperations;
import org.openide.filesystems.FileObject;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 *
 * @author Timon Veenstra
 */
@NbBundle.Messages({
    "FarmProject ID already registered=error opening project at {0}, a project with the same farm ID is already registered at path {1}",
    "FarmProject label warning=Warning!",
    "FarmProject warning multiple projects for farm id=Multiple projects detected for farm ID {0}/{1}",
    "FarmProject confirm forget previously registered project location=Do you want to forget about the project at {0}?"
})
public class FarmProject implements Project, PropertyChangeListener {

    /**
     * project type is used for registration of node factories like:
     * @NodeFactory.Registration(projectType = FarmProject.PROJECT_TYPE)
     */
    public static final String PROJECT_TYPE = "nl-cloudfarming-client-farm";
    private final FileObject projectDir;
    private final ProjectState state;
    private InstanceContent ic;
    private Lookup intern;
    private Lookup lkp;
    private final Farm farm;
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.farm");
    public static final String FARM_FILE = "farm.frm";
    public static final String ICON = "nl/cloudfarming/client/farm/house.png";
    private final FarmManagerImpl farmManager;

    /**
     *
     * @param dir
     * @param state
     */
    public FarmProject(FileObject dir, ProjectState state) {
        this.projectDir = dir;
        this.state = state;
        farmManager = Lookup.getDefault().lookup(FarmManagerImpl.class);
        this.farm = farmManager.load(this,dir.getFileObject(FARM_FILE));
        this.ic = new InstanceContent();
        this.intern = new AbstractLookup(this.ic);
        farm.addPropertyChangeListener(this);
        if (farm.getAddress() != null) {
            farm.getAddress().addPropertyChangeListener(this);
        }
    }

    @Override
    public FileObject getProjectDirectory() {
        return projectDir;
    }

    @Override
    public Lookup getLookup() {
        if (lkp == null) {
            lkp = Lookups.fixed(
                    state, //allow outside code to mark the project as needing saving
                    new ActionProviderImpl(), //Provides standard actions like Build and Clean
                    new FarmProjectDeleteOperation(),
                    new FarmProjectCopyOperation(this),
                    new FarmProjectInfo(), //Project information implementation
                    new FarmProjectLogicalView(this), //Logical view of project implementation
                    new FarmMoveOperation(this),
                    new FarmProjectOpenedHook(),
                    farmManager,
                    farm);
        }
        return new ProxyLookup(lkp, intern);
    }

    public Farm getFarm() {
        return farm;
    }
    
    public FarmID getFarmID() {
        return new FarmID(farm.getURI());
    }

    @Override
    public String toString() {
        return getProjectDirectory().getName();
    }


    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        if (intern.lookup(Savable.class) == null) {
            this.ic.add(new FarmSavable());
        }
    }
    
    private static final Icon IMAGE_ICON = ImageUtilities.loadImageIcon(ICON, true);

    private class FarmSavable extends EntitySavable {

        public FarmSavable() {
            super(farm, IMAGE_ICON);
            register();
        }

        @Override
        public void handleSave() throws IOException {
            project().farmManager.save(farm);
            destroy();
        }

        /**
         * Cancel the save operation Removes this Savable from the lookup and
         * unregisters
         */
        @Override
        public void destroy() {
            project().ic.remove(FarmSavable.this);
            unregister();
        }

        @Override
        protected String findDisplayName() {
            return farm.getName();
        }

        FarmProject project() {
            return FarmProject.this;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj instanceof FarmSavable) {
                return project() == ((FarmSavable) obj).project();
            }
            return false;
        }

        @Override
        public int hashCode() {
            return project().hashCode();
        }
    }

    private final class ActionProviderImpl implements ActionProvider {

        private String[] supported = new String[]{
            ActionProvider.COMMAND_DELETE,
            ActionProvider.COMMAND_COPY,};

        @Override
        public String[] getSupportedActions() {
            return supported;
        }

        @Override
        public void invokeAction(String string, Lookup lookup) throws IllegalArgumentException {
            if (string.equalsIgnoreCase(ActionProvider.COMMAND_DELETE)) {
                DefaultProjectOperations.performDefaultDeleteOperation(FarmProject.this);
            }
            if (string.equalsIgnoreCase(ActionProvider.COMMAND_COPY)) {
                DefaultProjectOperations.performDefaultCopyOperation(FarmProject.this);
            }
        }

        @Override
        public boolean isActionEnabled(String command, Lookup lookup) throws IllegalArgumentException {
            switch (command) {
                case ActionProvider.COMMAND_DELETE:
                    return true;
                case ActionProvider.COMMAND_COPY:
                    return true;
                default:
                    throw new IllegalArgumentException(command);
            }
        }
    }

    private final class FarmProjectDeleteOperation implements DeleteOperationImplementation {

        @Override
        public void notifyDeleting() throws IOException {
        }

        @Override
        public void notifyDeleted() throws IOException {
        }

        @Override
        public List<FileObject> getMetadataFiles() {
            List<FileObject> dataFiles = new ArrayList<>();
            return dataFiles;
        }

        @Override
        public List<FileObject> getDataFiles() {
            List<FileObject> dataFiles = new ArrayList<>();
            dataFiles.add(projectDir);
            return dataFiles;
        }
    }

    private final class FarmProjectCopyOperation implements CopyOperationImplementation {

        private final FarmProject project;
        private final FileObject projectDir;

        public FarmProjectCopyOperation(FarmProject project) {
            this.project = project;
            this.projectDir = project.getProjectDirectory();
        }

        @Override
        public List<FileObject> getMetadataFiles() {
            return Collections.EMPTY_LIST;
        }

        @Override
        public List<FileObject> getDataFiles() {
            return Collections.EMPTY_LIST;
        }

        @Override
        public void notifyCopying() throws IOException {
        }

        @Override
        public void notifyCopied(Project arg0, File arg1, String arg2) throws IOException {
        }
    }

    private final class FarmProjectInfo implements ProjectInformation {

        @Override
        public Icon getIcon() {
            return new ImageIcon(ImageUtilities.loadImage(ICON));
        }

        @Override
        public String getName() {
            return getProjectDirectory().getName();
        }

        @Override
        public String getDisplayName() {
            return getName();
        }

        @Override
        public void addPropertyChangeListener(PropertyChangeListener pcl) {
            //do nothing, won't change
        }

        @Override
        public void removePropertyChangeListener(PropertyChangeListener pcl) {
            //do nothing, won't change
        }

        @Override
        public Project getProject() {
            return FarmProject.this;
        }
    }

    private final class FarmMoveOperation implements MoveOrRenameOperationImplementation {

        private final FarmProject farmProject;

        public FarmMoveOperation(FarmProject farmProject) {
            this.farmProject = farmProject;
        }

        @Override
        public void notifyMoving() throws IOException {
        }

        @Override
        public void notifyMoved(Project original, File originalPath, String nueName) throws IOException {
        }

        @Override
        public List<FileObject> getMetadataFiles() {
            return ProjectOperations.getMetadataFiles(farmProject);
        }

        @Override
        public List<FileObject> getDataFiles() {
            return ProjectOperations.getDataFiles(farmProject);
        }

        @Override
        public void notifyRenaming() throws IOException {
        }

        @Override
        public void notifyRenamed(String nueName) throws IOException {
        }
    }
    
    private final class FarmProjectOpenedHook extends ProjectOpenedHook {
        @Override
        protected void projectOpened() {
            //check for unique farm ID across registered projects:
            final String currentProjectPath = FarmProject.this.getProjectDirectory().getPath();
            final String registeredPath = FarmProjectRegistry.getInstance().getRegisteredPath(FarmProject.this);
            
            if (registeredPath == null) {
                // new project, add to managed farms:
                FarmProjectRegistry.getInstance().add(FarmProject.this);
            }
            else if (!registeredPath.equals(currentProjectPath)) {
                // multiple farm project with same id, choose which one should be in the registry:
                LOGGER.log(Level.WARNING, Bundle.FarmProject_warning_multiple_projects_for_farm_id(farm.getIssuingAgency(), farm.getRegistrationNumber()));
                LOGGER.log(Level.INFO, Bundle.FarmProject_ID_already_registered(currentProjectPath, registeredPath));

                // ask if previously registered farm at different location should be forgotten:
                int option = JOptionPane.showConfirmDialog(
                        null,
                        Bundle.FarmProject_warning_multiple_projects_for_farm_id(farm.getIssuingAgency(), farm.getRegistrationNumber())
                        + "\n"
                        + Bundle.FarmProject_confirm_forget_previously_registered_project_location(registeredPath),
                        Bundle.FarmProject_label_warning(),
                        JOptionPane.YES_NO_OPTION);

                if (option == JOptionPane.YES_OPTION) {
                    // register the new location (overwrite entry in registry):
                    FarmProjectRegistry.getInstance().add(FarmProject.this);                    
                    // close the project at the registered location if it's currently open:
                    final FarmProject registeredProject = (FarmProject)findOpenProjectByPath(registeredPath);
                    
                    if (registeredProject != null) {
                        // close the project at the new location first, to ensure 
                        // that any open changes in the registered project will be
                        // saved in the right location:
                        OpenProjects.getDefault().close(new Project[]{FarmProject.this});
                        
                        // fire custom close action on registered project to handle savables:
                        List<FarmProject> list = Collections.singletonList(registeredProject);
                        new FarmProjectCloseAction(list).actionPerformed(null);
                        
                        boolean cancelled = OpenProjects.getDefault().isProjectOpen(registeredProject);
                        
                        if (cancelled) {
                            // rollback registry:
                            FarmProjectRegistry.getInstance().add(registeredProject); 
                        }
                        else {
                            OpenProjects.getDefault().open(new Project[]{FarmProject.this}, false);
                            // add new farm to cache (which was previously blocked by the instance from the registered project)
                            EntityCache.getInstance().add(farm);
                        }
                    }
                } else { // JOptionPane.NO_OPTION:
                    OpenProjects.getDefault().close(new Project[]{FarmProject.this});
                }
            }
        }

        @Override
        protected void projectClosed() {
            // clear entity cache for this farm:
            EntityCache.getInstance().remove(getFarmID());
        }
        
        private Project findOpenProjectByPath(final String path) {
            for (Project project : OpenProjects.getDefault().getOpenProjects()){
                if (project.getProjectDirectory().getPath().equals(path)) {
                    return project;
                }
            }
            return null;
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final FarmProject other = (FarmProject) obj;
        if (!Objects.equals(this.projectDir, other.projectDir)) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 17 * hash + Objects.hashCode(this.projectDir);
        return hash;
    }
}
